#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include <stdbool.h>
#include <ncurses.h>
#include <unistd.h>
#include <termios.h>
#include <limits.h>
#include <time.h>
#include <SDL2/SDL.h>
#include <SDL2/SDL_image.h>

#include "hash.h"
#include "structure.h"
#include "base.h"
#include "generation.h"
#include "display.h"
#include "move.h"

double speed = 0.5 ;
int precision = 12 ;

double epsilon_x = 0.0;
double epsilon_y = 0.0;

bool checkCollision() {
    return (unpack_coord(gridGet(map, player_cx, player_cy)->chdata.lines, player_x, player_y));
}

void increment_x(double dx) {
    epsilon_x = epsilon_x + dx ;
    if(epsilon_x >= 1.0) {
        epsilon_x -= 1.0 ;
        player_x += 1 ;
        if(player_x >= 8) {
            player_x -= 8 ;
            player_cx += 1;
        }
    } else if(epsilon_x < 0.0) {
        epsilon_x += 1.0 ;
        player_x -= 1 ;
        if(player_x < 0) {
            player_x += 8 ;
            player_cx -= 1;
        }
    }
}

void increment_y(double dy) {
    epsilon_y += dy ;
    if(epsilon_y >= 1.0) {
        epsilon_y -= 1.0 ;
        player_y += 1 ;
        if(player_y >= 8) {
            player_y -= 8 ;
            player_cy += 1;
        }
    } else if(epsilon_y < 0.0) {
        epsilon_y += 1.0 ;
        player_y -= 1 ;
        if(player_y < 0) {
            player_y += 8 ;
            player_cy -= 1;
        }
    }
}

bool moveRequest(double dx, double dy) {
    increment_x(dx);
    increment_y(dy);
    if(false && checkCollision()) {
        increment_x(-dx);
        increment_y(-dy);
        return false;
    };
    return true;
}

void moveControl(bool* halt) {
    SDL_Event event;
    int i = 0;
    while(SDL_PollEvent(&event)) {
        switch (event.type) {
            case SDL_QUIT:
                break;
            case SDL_KEYDOWN:
                switch (event.key.keysym.sym) {
                    case SDLK_z:
                        while(i < precision && moveRequest(0.0, -speed/precision)) {
                            i += 1;
                        };
                        break;
                    case SDLK_q:
                        while(i < precision && moveRequest(-speed/precision, 0.0)) {
                            i += 1;
                        };
                        break;
                    case SDLK_s:
                        while(i < precision && moveRequest(0.0, speed/precision)) {
                            i += 1;
                        };
                        break;
                    case SDLK_d:
                        while(i < precision && moveRequest(speed/precision, 0.0)) {
                            i += 1;
                        };
                        break;
                    case SDLK_TAB:
                        *halt = true;
                        break;
                    case SDLK_m:
                        zoom = max(25, zoom-25);
                        break;
                    case SDLK_p:
                        zoom = min(2250, zoom+25);
                        break;
                }
        }
    }
}

void moveFunctionMaster(SDL_Renderer* renderer) {
    bool halt = false ;
    while(!halt) {
        moveControl(&halt);
        resetRenderer(renderer);
        drawMapToRenderer(renderer, -300 * 250/zoom + to_int(50 * (8*player_cx + player_x + epsilon_x)), 300 * 250/zoom + to_int(50 * (8*player_cx + player_x + epsilon_x)), -300 * 250/zoom + to_int(50 * (8*player_cy + player_y + epsilon_y)), 300 * 250/zoom + to_int(50 * (8*player_cy + player_y + epsilon_y)));
        drawNumberToRenderer(renderer, digits, player_cx, 10, 10, 50, 70, 0);
        drawNumberToRenderer(renderer, digits, player_cy, 10, 80, 50, 70, 0);
        drawNumberToRenderer(renderer, digits, player_x, __width__ / 3, 10, 50, 70, 0);
        drawNumberToRenderer(renderer, digits, player_y, __width__ / 3, 80, 50, 70, 0);
        drawNumberToRenderer(renderer, digits, to_int(epsilon_x*100), 2 * __width__ / 3, 10, 50, 70, 0);
        drawNumberToRenderer(renderer, digits, to_int(epsilon_y*100), 2 * __width__ / 3, 80, 50, 70, 0);
        updateRenderer(renderer);
        draw_par += 1 ;
        usleep(20000);
    }
}